/**
  *  \file config.c
  *  \brief Config Parser
  *
  *  Configuration file parser. We understand a pconfig.src-alike
  *  file format.
  *
  *  \author Stefan Reuther <Streu@gmx.de>
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <stddef.h>
#include <limits.h>
#include "mfq.h"

static int parseBoolean(const struct ConfigOption*, const char* text, void* data);
static int parseLong(const struct ConfigOption*, const char* text, void* data);
static int parseLongArray(const struct ConfigOption*, const char* text, void* data);
static int parseScoreArray(const struct ConfigOption*, const char* text, void* data);

/** True iff there was an error. This variable is evaluated by the
    main function. */
Boolean config_had_errors;

/** True iff we read at least one option. This is used to detect
    the case where users run us with a `pconfig.src' but no mfq
    config inside. */
static Boolean config_file_read;

/** Global configuration. */
struct Config the_config;

#define CONF(name, type, default) { #name, offsetof(struct Config, name), parse##type, default }
struct ConfigOption config_parser[] = {
    CONF(ModifyMinefieldQuotas,     Boolean,     "Yes"),
    CONF(DeleteExcessMinefields,    Boolean,     "Yes"),
    CONF(UnmodifyMinefieldQuotas,   Boolean,     "No"),
    CONF(MinefieldCount,            Long,        "1000"),
    CONF(ScoreScaling,              LongArray,   "100"),
    CONF(ScorePerPlayer,            ScoreArray,  "0"),
    CONF(ScorePerPlanet,            ScoreArray,  "100+0"),
    CONF(ScorePerBase,              LongArray,   "100"),
    CONF(ScorePerCapitalShip,       ScoreArray,  "100+0"),
    CONF(ScorePerFreighter,         ScoreArray,  "10+0"),
    CONF(ScorePerMinefield,         ScoreArray,  "1+0"),
    CONF(AllowPlayerOverrides,      Boolean,     "Yes"),
};
#undef CONF

/** Assign to specified config option.
    \returns False on error. */
static Boolean
assignConfigOption(const struct ConfigOption* opt,
                   const char* rhs)
{
    void* data = ((char*)&the_config) + opt->offset;
    token_ptr = rhs;
    if (! opt->func(opt, rhs, data)) {
        config_had_errors = True;
        return False;
    } else {
        return True;
    }
}

/** Assign to config option given by name.
    \returns False on error. */
static Boolean
assignConfigStr(const char* lhs, char* rhs, const char* line)
{
    size_t i;
    if (!lhs)
        return True;

    for (i = 0; i < sizeof(config_parser)/sizeof(config_parser[0]); ++i) {
        if (StrEq(lhs, config_parser[i].name)) {
            config_file_read = True;
            return assignConfigOption(&config_parser[i], rhs);
        }
    }

    Error("Unknown config item `%s'", lhs);
    return False;
    UNUSED(line);
}

/** Load configuration. Exits with an error if not found. */
void
loadConfig(void)
{
    FILE* fp;
    size_t i;

    /* assign defaults */
    for (i = 0; i < sizeof(config_parser)/sizeof(config_parser[0]); ++i)
        assignConfigOption(&config_parser[i], config_parser[i].deflt);

    config_had_errors = False;
    config_file_read = False;

    /* look for this file only in the game directory, because pconfig.src
       is also looked-for only there. */
    if ((fp = OpenInputFile("mfq.cfg", GAME_DIR_ONLY | TEXT_MODE | NO_MISSING_ERROR))) {
        ConfigFileReader(fp, "mfg.cfg", "mfq", True, assignConfigStr);
        fclose(fp);
        if (!config_file_read)
            ErrorExit("Configuration information not found in `mfq.cfg'");
        return;
    }
    /* PHost looks for this one only in the game directory, so do we. */
    if ((fp = OpenInputFile("pconfig.src", GAME_DIR_ONLY | TEXT_MODE | NO_MISSING_ERROR))) {
        ConfigFileReader(fp, "pconfig.src", "mfq", False, assignConfigStr);
        fclose(fp);
        if (!config_file_read)
            ErrorExit("Configuration information not found in `pconfig.src'");
        return;
    }
    ErrorExit("Configuration information not found");
}

/** Parse a boolean. */
static int
parseBoolean(const struct ConfigOption* opt, const char* text, void* data)
{
    UNUSED(opt); UNUSED(text);

    if (getToken(1) != TT_ID)
        return False;
    if (StrEq(the_token, "Yes") || StrEq(the_token, "True"))
        *(Boolean*)data = True;
    else if (StrEq(the_token, "No") || StrEq(the_token, "False"))
        *(Boolean*)data = False;
    else
        return False;

    if (getToken(1) == TT_NULL)
        return True;
    else
        return False;
}

/** Parse a long. */
static int
parseLong(const struct ConfigOption* co, const char* text, void* data)
{
    UNUSED(co);

    token_ptr = text;
    if (getToken(1) != TT_INT)
        return False;
    *(long*)data = token_cooked.l;

    return getToken(1) == TT_NULL;
}

/** Parse an arrayized long option.
    The \c co parameter is not used; this is exploited by
    assignMinefieldQuotas below. */
static int
parseLongArray(const struct ConfigOption* co, const char* text, void* data)
{
    int copying = 0;
    int i, j;
    long* opt = data;

    UNUSED(co); UNUSED(text);

    for (i = 1; i <= 11; ++i) {
        if (copying) {
            opt[i] = opt[i-1];
        } else {
            if (getToken(1) != TT_INT)
                return False;
            opt[i] = token_cooked.l;
            j = getToken(1);
            if (j == TT_NULL)
                copying = 1;
            else if (j == ',')
                /* okay */;
            else
                return False;
        }
    }
    return getToken(1) == TT_NULL;
}

/** Parse a single ConfigScoreItem. Note that this reads the token
    after the item, too. */
static int
parseSingleScoreItem(struct ConfigScoreItem* opt)
{
    opt->base = 0;
    opt->mult = 0;
    opt->min = LONG_MIN;
    opt->max = LONG_MAX;

    /* first: base. This is mandatory. */
    if (getToken(1) != TT_INT)
        return False;
    opt->base = token_cooked.l;

    /* now the rest. They're all optional. */
    while (1) {
        switch(getToken(1)) {
         case '<':
            /* Maximum */
            if (getToken(1) != TT_INT)
                return False;
            opt->max = token_cooked.l;
            break;
         case '>':
            /* Minimum */
            if (getToken(1) != TT_INT)
                return False;
            opt->min = token_cooked.l;
            break;
         case '+':
            /* Add/Multiplier */
            switch (getToken(1)) {
             case TT_INT:
                opt->mult = token_cooked.l;
                break;
             case TT_FLOAT:
                opt->mult = token_cooked.d;
                break;
             default:
                return False;
            }
            break;
         default:
            return True;
        }
    }
}

/** Parse an array of ConfigScoreItem's. */
static int
parseScoreArray(const struct ConfigOption* co, const char* text, void* data)
{
    int copying = 0;
    int i, j;
    struct ConfigScoreItem* opt = data;

    UNUSED(text);

    for (i = 1; i <= 11; ++i) {
        if (copying) {
            opt[i] = opt[i-1];
        } else {
            if (!parseSingleScoreItem(&opt[i]))
                return False;
            if (opt[i].min > opt[i].max) {
                long l = opt[i].min;
                opt[i].min = opt[i].max;
                opt[i].max = l;
                Info("Minimum and maximum for `%s' in wrong order, swapping them", co->name);
            }
            j = token_type;
            if (j == TT_NULL)
                copying = 1;
            else if (j == ',')
                /* okay */;
            else
                return False;
        }
    }
    return getToken(1) == TT_NULL;
}

#if 0
/* The following is not actually needed because the PDK
   already parses it correctly ;) */

static long* amq_data;

/** Parse the MaximumMinefieldsPerPlayer option. */
static Boolean
assignMinefieldQuotas(const char* lhs, char* rhs, const char* line)
{
    if (lhs && StrEq(lhs, OPTION_NAME))
        return parseLongArray(0, token_ptr = rhs, amq_data);
    return True;
}

/** Load the PCONFIG.SRC, and extract the MaximumMinefieldsPerPlayer
    option. */
void
loadPHostMinefieldQuotas(long* data)
{
    FILE* fp;
    int i;

    for (i = 0; i <= 11; ++i)
        data[i] = 10000;

    fp = OpenInputFile("pconfig.src", GAME_DIR_ONLY | TEXT_MODE);
    amq_data = data;
    ConfigFileReader(fp, "pconfig.src", "phost", True, assignMinefieldQuotas);
    fclose(fp);
}
#endif
