/**
  *  \file delete.c
  *  \brief Minefield Deletion
  *
  *  This module contains functions for determining which minefields
  *  to delete, and to delete them.
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include "mfq.h"

/** Count minefields. Updates the count in data[]. */
static void
countMines(Uns16 data[])
{
    int mid, n;
    n = GetNumMinefields();
    for (mid = 1; mid <= n; ++mid)
        if (IsMinefieldExist(mid))
            ++data[MinefieldOwner(mid)];
}

/** Main entry point: delete excess mines. */
void
deleteMines(void)
{
    Uns16 mines_per_player[12];
    int i;

    Info("Deleting excess minefields...");
    for (i = 1; i <= 11; ++i)
        mines_per_player[i] = 0;
    countMines(mines_per_player);

    for (i = 1; i <= 11; ++i)
        if (mines_per_player[i] > player_quotas[i])
            deleteMinesOfPlayer(i, mines_per_player[i],
                                mines_per_player[i] - player_quotas[i]);
}

/**********************************************************************/

/** Information about a player's minefield. */
struct MinefieldInfo {
    Uns16 mine_id;              /**< Id of this minefield. */
    Uns32 distance;             /**< Minimum distance to key point, squared. */
};

/** Pointer to an array of minefield info structures for this player,
    num_mines in total.

    The first num_fixed minefields were explicitly "promoted" using
    explicit commands (xtern keep). */
static struct MinefieldInfo* this_mines;
static int num_mines;
static int num_fixed;

static void performPlayerOverrides(int player);

/** Delete mines for one player.
    \param player player Id
    \param have   how many minefields this player has
    \param count  number of minefields to delete */
void
deleteMinesOfPlayer(Uns16 player, int have, int count)
{
    int i, n;
    if (flag_verbose)
        Info("    Player %d has %d minefields over limit", player, count);

    /* Fill the minefield list. */
    this_mines = MemAlloc(sizeof(struct MinefieldInfo) * have);
    num_fixed = num_mines = 0;
    n = GetNumMinefields();
    for (i = 1; i <= n; ++i) {
        if (IsMinefieldExist(i) && MinefieldOwner(i) == player) {
            this_mines[num_mines].mine_id = i;
            this_mines[num_mines].distance = ~0UL;
            ++num_mines;
            passert(num_mines <= have);
        }
    }
    passert(num_mines == have);

/*     for (i = 0; i < num_mines; ++i) */
/*         printf(" %d", this_mines[i].mine_id); */
/*     printf("\n"); */
    performPlayerOverrides(player);
/*     for (i = 0; i < num_mines; ++i) */
/*         printf(" %d", this_mines[i].mine_id); */
/*     printf("\n"); */

    for (i = 1; i <= count; ++i) {
        int mid = this_mines[num_mines - i].mine_id;
        if (flag_verbose) {
            Uns32 dist = this_mines[num_mines - i].distance;
            int r = MinefieldRadius(mid);
            if (dist != ~0UL)
                Info("        Deleting %d (radius %d, distance %lu)", mid, r, dist);
            else
                Info("        Deleting %d (radius %d)", mid, r);
        }
        passert(MinefieldOwner(mid) == player);
        PutMinefieldUnits(mid, 0);
        PutMinefieldOwner(mid, 0);
    }

    MemFree(this_mines);
}

/** Perform one minefield id override (`keep #id' command). We do
    that by moving the specified id to the head of the list. */
static void
performMineIdOverride(int mid)
{
    int i, j;
    for (i = num_fixed; i < num_mines; ++i) {
        if (this_mines[i].mine_id == mid) {
            struct MinefieldInfo mf = this_mines[i];
            for (j = i; j > num_fixed; --j)
                this_mines[j] = this_mines[j-1];
            this_mines[num_fixed++] = mf;
            return;
        }
    }
}

/** Perform X/Y override (`keep x,y' command). Updates the distance
    in all minefields that are not yet "fixed" by a minefield id
    override. */
static void
performXYOverride(int x, int y)
{
    int i;
    for (i = num_fixed; i < num_mines; ++i) {
        Uns32 dist = DistanceSquared(x, y,
                                     MinefieldPositionX(this_mines[i].mine_id),
                                     MinefieldPositionY(this_mines[i].mine_id));
        if (dist < this_mines[i].distance)
            this_mines[i].distance = dist;
    }
}

/** Compare minefields by importance (qsort comparison).
    A minefield is more important than another if it is nearer to a
    key point, or larger than another one. If in doubt, compares by Id,
    just to get it deterministic. */
static int
compareMines(const void* a, const void* b)
{
    const struct MinefieldInfo* ma = a;
    const struct MinefieldInfo* mb = b;
    Uns32 ua, ub;

    if (ma->distance < mb->distance)
        return -1;
    else if(ma->distance > mb->distance)
        return +1;

    ua = MinefieldUnits(ma->mine_id);
    ub = MinefieldUnits(mb->mine_id);
    if (ua > ub)
        return -1;
    else if (ua < ub)
        return +1;

    return ma->mine_id - mb->mine_id;
}

/** Perform all overrides for the specified player.
    \pre this_mines points to num_mines minefields, num_fixed is
    usually zero
    \post this_mines is sorted so that the most important minefields
    (those we do not want to delete) are first. */
static void
performPlayerOverrides(int player)
{
    struct PlayerOverride* p = first;
    while (p) {
        if (p->player == player) {
            if (p->is_mineid)
                performMineIdOverride(p->x);
            else
                performXYOverride(p->x, p->y);
        }
        p = p->next;
    }

    if (num_fixed < num_mines)
        qsort(&this_mines[num_fixed], num_mines - num_fixed,
              sizeof(struct MinefieldInfo), compareMines);
}
