/**
  *  \file mfq.c
  *  \brief Main program
  *
  *  Command line parsing, overall control.
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <stdio.h>
#include <stdarg.h>
#include "mfq.h"

static int phase = 1;
static const char* progname;

int flag_readonly = 0;
int flag_verbose = 0;

/** Friendly help message. */
static void
help()
{
    printf("MFQ -- Minefield Quotas for PHost 3.3 -- Version " VERSION "\n\n"
           "Usage: %s [options] [gamedir [rootdir]]\n\n"
           "-h -H\tthis friendly help message\n"
           "-V\tprint version number\n"
           "-v\tverbose\n"
           "-r\tread-only operation\n"
           "-1\tdo phase 1 (AUXHOST1)\n"
           "-2\tdo phase 2 (AUXHOST2) [see docs!]\n\n"
           COPYRIGHT "\n",
           progname);
    exit(0);
}

/** Version information. */
static void
version()
{
    printf("MFQ -- Minefield Quotas for PHost 3.3 -- Version " VERSION "\n");
}

static void
showScores()
{
    int i;
    Info("Scores");
    Info("          Player               Score      Minefields");
    Info("       -- -------------------- ---------- ----------");
    for (i = 1; i <= 11; ++i)
        if (PlayerIsActive(i))
            Info("       %2d %-20s %10d %10d",
                 i, RaceShortName(i, 0), player_scores[i], player_quotas[i]);
    Info("");
}

/** Do phase 1. Phase 1 contains:
    - score computation
    - modification of the pconfig.src file
    - expiring minefields */
static void
doPhase1()
{
    int i;

    Info("=== MFQ Phase 1 ===");
    Info("Loading game data...");
    if (! ReadGlobalData() || ! ReadHostData())
        exit(1);

    if (the_config.ModifyMinefieldQuotas) {
        computeScores();
        computeQuotas();
        if (flag_verbose)
            showScores();
        modifyPConfig(player_quotas);
    } else {
        Info("Using minefield quotas from pconfig.src");
        for (i = 1; i <= 11; ++i)
            player_quotas[i] = gPconfigInfo->MaximumMinefieldsPerPlayer[i];
    }

    if (the_config.DeleteExcessMinefields) {
        if (the_config.AllowPlayerOverrides)
            loadPlayerOverrides();
        deleteMines();
        if (!flag_readonly) {
            Info("Saving game data...");
            if (!Write_Mines_File())
                Error("Error writing the mines file.");
        } else
            Info("Read-only operation: not saving game data.");
    }

    if (the_config.UnmodifyMinefieldQuotas)
        saveStatus(gPconfigInfo->MaximumMinefieldsPerPlayer);
}

/** Do phase 2. Phase 2 restores the MaximumMinefieldsPerPlayer
    option if needed. */
static void
doPhase2()
{
    Uns16 data[12];

    Info("=== MFQ Phase 2 ===");
    if (!the_config.UnmodifyMinefieldQuotas) {
        Info("Phase 2 not needed -- exiting...");
        return;
    }
    loadStatus(data);
    modifyPConfig(data);
}

int
main(int argc, char** argv)
{
    /* Process Options */
    int c;
    progname = argv[0];
    while((c = pgetopt(argc, argv, "hHvVr12")) != EOF) {
        switch(c) {
         case 'h':
         case 'H':
            help();
            exit(0);
         case 'V':
            version();
            exit(0);
         case 'v':
            flag_verbose = 1;
            break;
         case 'r':
            flag_readonly = 1;
            break;
         case '1':
            phase = 1;
            break;
         case '2':
            phase = 2;
            break;
         default:
            exit(1);
        }
    }
    if (poptind < argc)
        gGameDirectory = argv[poptind++];
    if (poptind < argc)
        gRootDirectory = argv[poptind++];
    if (poptind < argc)
        ErrorExit("too many parameters");

    /* Actual Work */
    if (phase != 2)
        InitPHOSTLib();        /* we need the PDK only in phase 1 */
    loadConfig();
    if (config_had_errors)
        ErrorExit("Configuration file had errors, exiting.");
    if (!the_config.ModifyMinefieldQuotas && !the_config.DeleteExcessMinefields)
        ErrorExit("You have not selected a task to perform in the config file.");

    if (phase == 1)
        doPhase1();
    else
        doPhase2();

    return 0;
}
