/**
  *  \file override.c
  *  \brief Player Override Management
  *
  *  This module reads and parses player overrides into a list.
  *  The delete.c module can use that to control deletion of
  *  minefields.
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <stdio.h>
#include "mfq.h"

struct PlayerOverride* first = 0;
struct PlayerOverride* last = 0;

static char* the_line;
static Boolean player_got_intro[12];

/** Add override. Appends a new override structure to the end of the
    list.
    \param player     player this override belongs to
    \param is_mineid  True iff this override specifies a minefield Id
    \param x,y        x/y coordinates if is_mineid is False;
                      x is a minefield Id if is_mineid is True. */
static void
addOverride(int player, Boolean is_mineid, int x, int y)
{
    struct PlayerOverride* p = MemAlloc(sizeof(struct PlayerOverride));

    p->next      = 0;
    p->player    = player;
    p->is_mineid = is_mineid;
    p->x         = x;
    p->y         = y;

    if (first)
        last->next = p;
    else
        first = p;
    last = p;

    if (flag_verbose) {
        if (is_mineid)
            Info("        keep #%d", x);
        else
            Info("        keep %d,%d", x, y);
    }
}

/** Read one line from file. Returns MemAlloc'd buffer, or NULL on
    error. */
static char*
getline(FILE* fp)
{
    static char buffer[4096];
    char* data = 0;
    size_t data_len = 0;        /* length of the string data */

    while (fgets(buffer, sizeof buffer, fp)) {
        size_t len = strlen(buffer);
        data = MemRealloc(data, data_len + len + 1);
        strcpy(data + data_len, buffer);
        data_len += len;
        if (data_len && data[data_len-1] == '\n') {
            data[data_len-1] = 0;
            break;
        }
    }
    return data;
}

/** Send introductory message. Sends an explanation of the command
    syntax; as response to the `help' command. */
static void
sendPlayerIntro(int player)
{
    if (player <= 0 || player > 11)
        return;

    if (!player_got_intro[player]) {
        player_got_intro[player] = True;
        if (! flag_readonly)
            WriteAUXHOSTMessage(player,
                                "(-h0000)<<< MFQ Command Help >>>\015"
                                "\015"
                                "Valid command forms:\015"
                                "  keep #mid\015"
                                "  keep x,y\015"
                                "  help\015"
                                "where mid is a minefield id\015"
                                "x,y is a coordinate pair.\015"
                                "\015"
                                "Either write those into a file\015"
                                "`mfq%d.txt' and submit that with your\015"
                                "TRN, or use the command processor and\015"
                                "send a sub-space message containing\015"
                                "  xtern <command>\015"
                                "to yourself.\015");
    }
}

/** Report invalid command to player.
    \param player  evildoer
    \param comment short notice why the command was invalid. */
static void
reportInvalidCommand(int player, const char* comment)
{
    char buf[1000];
    if (player <= 0 || player > 11)
        return;

    if (flag_readonly) {
        Info("Player %d, command `%s': %s", player, the_line, comment);
    } else {
        sprintf(buf,
                "(-h0000)<<< MFQ Error >>>\015"
                "\015"
                "The following command\015"
            "  %.38s\015"
                "could not be executed:\015"
                "  %s\015",
                the_line, comment);
        WriteAUXHOSTMessage(player, buf);
    }

    sendPlayerIntro(player);
}

/** Parse command.
    \param player  the player this command belongs to
    \param allow_invalid_commands if nonzero, unknown commands are not
                   reported as errors. This is used for xterncmd.ext
                   parsing; that file might contain commands for other
                   add-ons. */
static void
parseCommand(int player, int allow_invalid_commands)
{
    int t = getToken(0);
    if (t == TT_NULL)
        return;                 /* blank line or comment */
    if (t != TT_ID) {
        if (!allow_invalid_commands)
            reportInvalidCommand(player, "Invalid command word");
        return;
    }
    if (StrEq(the_token, "keep")) {
        /* It is a `keep' command. Syntactic forms we allow:
              keep x y
              keep x,y
              keep #id */
        int x, y;
        switch(getToken(0)) {
         case '#':
            if (getToken(0) != TT_INT || token_cooked.l <= 0 || token_cooked.l >= GetNumMinefields()) {
                reportInvalidCommand(player, "Invalid minefield Id");
                return;
            }
            x = token_cooked.l;
            if (getToken(0) != TT_NULL)
                reportInvalidCommand(player, "Trailing garbage");
            else
                addOverride(player, True, x, 0);
            break;
         case TT_INT:
            x = token_cooked.l;
            if (x < 0 || x > MAX_COORDINATE) {
                reportInvalidCommand(player, "Invalid X coordinate");
                return;
            }
            if ((t = getToken(0)) == ',')
                t = getToken(0);
            if (t != TT_INT || token_cooked.l < 0 || token_cooked.l > MAX_COORDINATE) {
                reportInvalidCommand(player, "Invalid Y coordinate");
                return;
            }
            y = token_cooked.l;
            if (getToken(0) != TT_NULL)
                reportInvalidCommand(player, "Trailing garbage");
            else
                addOverride(player, False, x, y);
            break;
         default:
            reportInvalidCommand(player, "Invalid parameter to `keep'");
        }
    } else if (StrEq(the_token, "help")) {
        sendPlayerIntro(player);
    } else {
        if (!allow_invalid_commands)
            reportInvalidCommand(player, "Invalid command word");
    }
}

/** Load xterncmd.ext file. */
static void
loadXternFile(const char* fname)
{
    FILE* fp;
    int player;

    fp = OpenInputFile(fname, TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (!fp)
        return;
    Info("    %s", fname);

    /* Skip first line */
    MemFree(getline(fp));

    /* Read rest of file */
    while ((the_line = getline(fp)) != 0) {
        token_ptr = the_line;
        /* Line starts with `playerid:' */
        if (getToken(0) != TT_INT)
            goto next;
        player = token_cooked.l;
        if (getToken(0) != ':')
            goto next;
        parseCommand(player, 1);
     next:
        MemFree(the_line);
    }

    fclose(fp);
}

/** Load mfqX.txt file.
    \param player player number
    \param fname  file name ("mfq<player>.txt") */
static void
loadMfqFile(int player, const char* fname)
{
    FILE* fp;

    fp = OpenInputFile(fname, TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (!fp)
        return;

    Info("    %s", fname);
    while ((the_line = getline(fp)) != 0) {
        token_ptr = the_line;
        parseCommand(player, 0);
        MemFree(the_line);
    }

    fclose(fp);
}

/** Main entry point: load player overrides. This loads the xterncmd.ext
    file, as well as the individual mfqX.txt files. */
void
loadPlayerOverrides(void)
{
    int i;
    char buf[20];

    Info("Loading player overrides...");
    loadXternFile("xterncmd.ext");
    for (i = 1; i <= 11; ++i) {
        if (PlayerIsActive(i)) {
            sprintf(buf, "mfq%d.txt", i);
            loadMfqFile(i, buf);
        }
    }
}
