/**
  *  \file score.c
  *  \brief Score computation
  *
  *  This file contains functions to compute the score, and the
  *  minefield quotas.
  *
  *  \author Stefan Reuther <Streu@gmx.de>
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */

#include <limits.h>
#include <stdlib.h>
#include "mfq.h"

#define MAX_PLAYER_QUOTA 10000

/** Scores for each player. Indexed by player number. */
long player_scores[12];

/** Player quotas. Indexed by player number. This must have the
    same type as gPconfigInfo->MaximumMinefieldsPerPlayer */
Uns16 player_quotas[12];

/** Add to score. This function attempts to prevent overflow. */
static void
addScore(int player, long score)
{
    if (score > 0 && score > LONG_MAX - player_scores[player])
        player_scores[player] = LONG_MAX;
    else if (score < 0 && player_scores[player] < LONG_MIN - score)
        player_scores[player] = LONG_MIN;
    else
        player_scores[player] += score;
}

/** Compute the score from a ConfigScoreItem, using parameter arg. */
static long
computeScore(const struct ConfigScoreItem* ci, long arg)
{
    long l = ci->base;
    if (arg && ci->mult) {
        double d = l + ci->mult * arg;
        if (d < LONG_MIN)
            l = LONG_MIN;
        else if (d > LONG_MAX)
            l = LONG_MAX;
        else
            l = d;
    }
    if (l < ci->min)
        return ci->min;
    else if (l > ci->max)
        return ci->max;
    else
        return l;
}

/** Compute scores for planets and starbases. */
static void
computePlanetScores(void)
{
    int pid;
    RaceType_Def player;

    Info("    planets...");
    for (pid = 1; pid <= PLANET_NR; ++pid) {
        if (!IsPlanetExist(pid))
            continue;
        if (! (player = PlanetOwner(pid)))
            continue;

        /* okay, planet exists. */
        addScore(player, computeScore(&the_config.ScorePerPlanet[player],
                                      PlanetCargo(pid, COLONISTS) / 100));
        if (IsBaseExist(pid))
            addScore(player, the_config.ScorePerBase[player]);
    }
}

/** Compute scores for ships. */
static void
computeShipScores(void)
{
    int sid;
    RaceType_Def player;
    int mass;

    Info("    starships...");
    for (sid = 1; sid <= SHIP_NR; ++sid) {
        if (!IsShipExist(sid))
            continue;
        player = ShipOwner(sid);

        /* okay, ship exists. */
        mass = HullMass(ShipHull(sid));
        if (ShipBeamNumber(sid) || ShipTubeNumber(sid) || ShipBays(sid))
            addScore(player,
                     computeScore(&the_config.ScorePerCapitalShip[player], mass));
        else
            addScore(player,
                     computeScore(&the_config.ScorePerFreighter[player], mass));
    }
}

/** Compute scores for minefields. */
static void
computeMinefieldScores()
{
    int mid;
    RaceType_Def player;

    Info("    minefields...");
    for (mid = 1; mid <= GetNumMinefields(); ++mid) {
        if (!IsMinefieldExist(mid))
            continue;
        player = MinefieldOwner(mid);

        addScore(player, computeScore(&the_config.ScorePerMinefield[player],
                                      MinefieldUnits(mid)));
    }
}

/** Main entry point: compute scores. */
void
computeScores(void)
{
    int i;
    Info("Computing scores:");
    for (i = 0; i <= 11; ++i)
        player_scores[i] = 0;
    for (i = 1; i <= 11; ++i)
        addScore(i, computeScore(&the_config.ScorePerPlayer[i],
                                 GetActivityLevel(i)));

    computePlanetScores();
    computeShipScores();
    computeMinefieldScores();
}

/**********************************************************************/

/** Utility function: sort players by their score. */
static int
sortByQuota(const void* a, const void* b)
{
    long qa = player_scores[*(const int*)a];
    long qb = player_scores[*(const int*)b];
    if (qa < qb)
        return -1;
    else if (qa > qb)
        return 1;
    else
        return 0;
}

/** Main entry point: Compute quotas.
    \pre computeScores() must have been called. */
void
computeQuotas(void)
{
    int i;
    double scores[12], total_score;
    long slots;                 /* remaining slots */
    int sort[12];               /* players by score, ascending */
    int qptr;
    int skip_counter;

    /* Compute the total score we want to distribute */
    total_score = 0;
    for (i = 1; i <= 11; ++i) {
        if (PlayerIsActive(i)) {
            scores[i] = player_scores[i] * (double)the_config.ScoreScaling[i];
            if(scores[i] < 0)
                scores[i] = 0;
            total_score += scores[i];
        } else
            scores[i] = 0;
    }

    /* Assign slots to players. This has rounding errors and
       "loses" some slots. */
    slots = the_config.MinefieldCount;
    if (slots <= 0)
        ErrorExit("Config error: MinefieldCount must be positive");

    player_quotas[0] = 0;
    if (total_score == 0) {
        /* pathological case */
        total_score = 11;
        for (i = 1; i <= 11; ++i)
            scores[i] = 1;
    }

    for (i = 1; i <= 11; ++i) {
        long l = scores[i] / total_score * the_config.MinefieldCount;
        if (l > MAX_PLAYER_QUOTA)
            l = MAX_PLAYER_QUOTA;
        player_quotas[i] = l;
        slots -= l;
    }

    /* there are some slots remaining. Distribute them. */
    for (i = 1; i <= 11; ++i)
        sort[i] = i;
    qsort(&sort[1], 11, sizeof (int), sortByQuota);
    qptr = 0;

    skip_counter = 0;
    while (slots && skip_counter < 11) {
        if (slots > 0) {
            /* first give those with the lowest score. */
            if (++qptr > 11)
                qptr = 1;
            i = sort[qptr];
            if (PlayerIsActive(i) && player_quotas[i] < MAX_PLAYER_QUOTA)
                player_quotas[i]++, slots--, skip_counter=0;
            else
                ++skip_counter;
        } else {
            /* can this happen? (too many slots distributed initially)
               I think no, but better safe than sorry. */
            if (--qptr <= 0)
                qptr = 11;
            i = sort[qptr];
            if (PlayerIsActive(i) && player_quotas[i])
                player_quotas[i]--, slots++, skip_counter=0;
            else
                ++skip_counter;
        }
    }
    if (slots) {
        Warning("%ld slots were not allocated.", slots);
        Warning("  This can happen when you set `MinefieldCount' far too high, or there");
        Warning("  too few players in your game.");
    }
}
