/**
  *  \file mfq.h
  *  \brief Main Include File
  *
  *  This file contains prototypes for all functions etc. in this
  *  program.
  *
  *  (c) copyright 2001 Stefan Reuther, Thomas Klebes.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU General Public License
  *  version 2, as published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */
#ifndef MFQ_H_INCLUDED
#define MFQ_H_INCLUDED

#include <stddef.h>
#include <phostpdk.h>
#include <limits.h>

#define UNUSED(x) (void)&x

#define VERSION "0.1.1"
#define COPYRIGHT "(c) 2001 Stefan Reuther, Thomas Klebes"

struct ConfigOption;

typedef int (*ConfigFunction)(const struct ConfigOption*,
                              const char* text,
                              void* data);

struct ConfigOption {
    const char*    name;        /**< Name of the config option */
    ptrdiff_t      offset;      /**< Offset in the configuration structure */
    ConfigFunction func;        /**< Function that parses this item */
    const char*    deflt;       /**< Default value */
};

struct ConfigScoreItem {
    long   base;
    double mult;
    long   min;
    long   max;
};

struct Config {
    /** If enabled, mfq computes per-player minefield quotas and
        manipulates pconfig.src accordingly. This is done in AUXHOST1. */
    Boolean ModifyMinefieldQuotas;

    /** If enabled, mfq deletes minefields above the pconfig.src limit
        (MinefieldsPerRace). This is done in AUXHOST1. */
    Boolean DeleteExcessMinefields;

    /** If enabled, mfq restores the per-player minefield quotas to
        their previous values in AUXHOST2, so players do not see them. */
    Boolean UnmodifyMinefieldQuotas;

    /** Number of minefield slots we're distributing among the players.
        Can (should!) be more than configured in pconfig.src
        (CPNumMinefields). For example, when this is 1100, each player
        will be granted 100 minefields. Assuming CPNumMinefields=500:
        when there are only 5 active mine-layers, they can completely
        exploit that limit; when all 11 players lay minefields, they'll
        hit the CPNumMinefield limit before their individual limit. */
    long MinefieldCount;

    /** Per-race score scaling (arrayized) */
    long ScoreScaling[12];

    /**  Scores. All options are arrayized. */
    struct ConfigScoreItem ScorePerPlayer[12];
    struct ConfigScoreItem ScorePerPlanet[12];
    long ScorePerBase[12];
    struct ConfigScoreItem ScorePerCapitalShip[12];
    struct ConfigScoreItem ScorePerFreighter[12];
    struct ConfigScoreItem ScorePerMinefield[12];

    /** Allow players to override which minefields they want to
        delete. */
    Boolean AllowPlayerOverrides;
};

/* config.c */
#define OPTION_NAME "MaximumMinefieldsPerPlayer"
extern struct Config the_config;
extern struct ConfigOption config_parser[];
extern Boolean config_had_errors;

void loadConfig(void);
/* void loadPHostMinefieldQuotas(long* data); */

/* util.c */
int StrEq(const char* src, const char* dest);

/* tok.c */
union TokenVal {
    double d;
    long l;
};
#define TT_NULL  (CHAR_MAX + 1)
#define TT_INT   (CHAR_MAX + 2)
#define TT_FLOAT (CHAR_MAX + 3)
#define TT_ID    (CHAR_MAX + 4)
#define TT_ERROR (CHAR_MAX + 5)
extern char the_token[];
extern int token_type;
extern union TokenVal token_cooked;
extern const char* token_ptr;
int getToken(int reading_config);

/* score.c */
extern long player_scores[12];
extern Uns16 player_quotas[12];
void computeScores(void);
void computeQuotas(void);

/* tmpfile.c */
void saveStatus(Uns16 data[]);
void loadStatus(Uns16 data[]);

/* modify.c */
void modifyPConfig(Uns16 data[]);

/* delete.c */
void deleteMines(void);
void deleteMinesOfPlayer(Uns16 player, int have, int count);

/* mfq.c */
extern int flag_readonly, flag_verbose;

/* override.c */
struct PlayerOverride {
    struct PlayerOverride* next; /**< Link */
    int player;                  /**< Player number */
    Boolean is_mineid;           /**< True iff this override explicitly names a minefield */
    int x, y;                    /**< x,y coordinates, or x = mine id */
};
struct PlayerOverride* first;
struct PlayerOverride* last;

void loadPlayerOverrides(void);

#endif
